package org.bitbucket.bradleysmithllc.etlunit.feature.file;

import com.google.inject.name.Named;
import org.apache.commons.io.FileUtils;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.RuntimeOption;
import org.bitbucket.bradleysmithllc.etlunit.feature.extend.Extender;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file._assert.AssertHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file._assert.AssertRequest;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.io.file.*;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataArtifact;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataArtifactContent;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataContext;
import org.bitbucket.bradleysmithllc.etlunit.metadata.MetaDataPackageContext;
import org.bitbucket.bradleysmithllc.etlunit.metadata.impl.MetaDataImpl;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;
import org.bitbucket.bradleysmithllc.json.validator.ResourceNotFoundException;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;
import java.util.*;

public class FileAssertExtender implements Extender, AssertHandler
{
	private final FileFeatureModule fileFeatureModule;
	private FileRuntimeSupport fileRuntimeSupport;
	private RuntimeSupport runtimeSupport;

	private DiffManager diffManager;
	private Log applicationLog;

	private DataFileManager dataFileManager;

	private boolean refreshAssertionData = false;

	public FileAssertExtender(FileFeatureModule fileFeatureModule)
	{
		this.fileFeatureModule = fileFeatureModule;
	}

	@Inject
	public void receiveDataFileManager(DataFileManager dfm)
	{
		dataFileManager = dfm;
	}

	@Inject
	public void receiveRuntimeSupport(RuntimeSupport rs)
	{
		runtimeSupport = rs;
	}

	@Inject
	public void receiveApplicationLog(@Named("applicationLog") Log log)
	{
		this.applicationLog = log;
	}

	@Inject
	public void setRefreshAssertionData(@Named("file.refreshAssertionData") RuntimeOption option)
	{
		refreshAssertionData = option.isEnabled();

		applicationLog.info(refreshAssertionData ? "File Module refreshing assertion data" : "File Module not refreshing assertion data");
	}

	@Inject
	public void receiveFileRuntimeSupport(FileRuntimeSupport support)
	{
		fileRuntimeSupport = support;
	}

	@Inject
	public void receiveDiffManager(DiffManager reporter)
	{
		diffManager = reporter;
	}

	public Feature getFeature()
	{
		return fileFeatureModule;
	}

	public action_code _assert(AssertRequest operation, ETLTestMethod mt, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		MetaDataContext fcontext = fileFeatureModule.getDataMetaContext();

		MetaDataPackageContext pcontext = fcontext.createPackageContextForCurrentTest(MetaDataPackageContext.path_type.test_source);

		String fileName = operation.getFile();

		File assertionSource = fileRuntimeSupport.getAssertionFileForCurrentTest(fileName);
		MetaDataArtifact pcontextArtifact;
		MetaDataArtifactContent artContent;

		String targetFileName = operation.getTargetFileName();

		if (targetFileName == null)
		{
			targetFileName = fileName;
		}

		FileProducer fp = fileRuntimeSupport.getRegisteredProducer(operation.getProducer());

		File target = fp.locateFileWithClassifierAndContext(targetFileName, operation.getClassifier(), operation.getContext());

		// check for the operation type:
		AssertRequest.AssertionMode assertionMode = operation.getAssertionMode();

		if (assertionMode == null)
		{
			assertionMode = AssertRequest.AssertionMode.EQUALS;
		}

		switch (assertionMode)
		{
			case EQUALS:
				// this is the usual case and is handled outside this switch
				break;
			case EXISTS:
				// target file must exist.  It's a failure if it does not.
				if (!target.exists())
				{
					String failureId = FileConstants.FAIL_FILE_NOT_EXIST;

					if (operation.getFailureId() != null)
					{
						failureId = operation.getFailureId();
					}

					throw new TestAssertionFailure(target.getAbsolutePath(), failureId);
				}

				return ClassResponder.action_code.handled;
			case NOT_EXISTS:
				// target file must not exist.  It's a failure if it does.
				if (target.exists())
				{
					String failureId = FileConstants.FAIL_FILE_EXISTS;

					if (operation.getFailureId() != null)
					{
						failureId = operation.getFailureId();
					}

					throw new TestAssertionFailure(target.getAbsolutePath(), failureId);
				}

				return ClassResponder.action_code.handled;
			case EMPTY:
				// target file must exist and be empty.  It's a failure if it does not exist or contains data.
			case NOT_EMPTY:
				// target file must exist and contain data.  It's a failure if it does not exist or does not contain data.
				simpleCheckContents(target, obj, operation, assertionMode, op, context);

				return ClassResponder.action_code.handled;
		}

		AssertRequest.ColumnListMode columnListMode = operation.getColumnListMode();

		List<String> actualList = new ArrayList<String>();
		Set<String> list = operation.getColumnList();

		if (columnListMode == null)
		{
			columnListMode = AssertRequest.ColumnListMode.EXCLUDE;
		}
		else
		{
			if (list == null)
			{
				throw new TestExecutionError("Must specify column-list when the column-list-mode attribute is speicified", FileConstants.ERR_STATE_COLUMN_MODE_STATE);
			}
		}

		/** It's okay for the source file not to exist if refresh assertion data is enabled
		 *  since the source file will be refreshed prior to the assertion being tested
		 */
		if (!assertionSource.exists() && !refreshAssertionData)
		{
			throw new TestExecutionError("Target file does not exist: " + assertionSource, FileConstants.ERR_TARGET_FILE_MISSING);
		}

		if (target == null || !target.exists())
		{
			throw new TestExecutionError("Source file does not exist: " + target, FileConstants.ERR_SOURCE_FILE_MISSING);
		}

		try {
			pcontextArtifact = pcontext.createArtifact("data", fileRuntimeSupport.getAssertionFileDirForCurrentTest());
			artContent = pcontextArtifact.createContent(assertionSource.getName());
			artContent.referencedByCurrentTest("assertion-source-file");
		} catch (IOException e) {
			throw new TestExecutionError("Error creating meta data artifact", e);
		}

		DataFileSchema ffs = null;

		try
		{
			ffs = fileRuntimeSupport.locateReferenceFileSchemaForCurrentTest(operation.getFile(), null, obj);

			applicationLog.info("Resolved data file schema: " + ffs.getId());

			context.declareAndSetStringValue(FileConstants.ASSERT_FILE_SCHEMA, ffs.getId());
		}
		catch (IllegalArgumentException exc)
		{
			throw new TestExecutionError("Error loading Flat File Schema", FileConstants.ERR_LOAD_FFML, exc);
		}
		catch (ResourceNotFoundException exc)
		{
			throw new TestExecutionError("Error resolving Flat File Schema", FileConstants.ERR_LOAD_FFML, exc);
		}

		actualList.addAll(ffs.getColumnNames());

		if (list != null)
		{
			switch (columnListMode)
			{
				case INCLUDE:
					// our list only includes mentioned columns
					actualList.clear();
					actualList.addAll(list);
					break;
				case EXCLUDE:
					// start with all columns, and remove the ones that are called out
					actualList.removeAll(list);
					break;
			}
		}

		// create a delimited local copy of the remote file
		String localID = op.getQualifiedName() + "-target-" + ffs.getId() + "." + target.getName() + ".delimited";

		File localExpected = assertionSource;
		File localTarget = fileRuntimeSupport.getGeneratedDataFileForCurrentTest(localID);

		try
		{
			DataFileSchema ffView = ffs;

			ffView = ffs.createSubViewIncludingColumns(actualList, "synthetic-" + ffs.getId() + "-delimited-" + op.getQualifiedName(), DataFileSchema.format_type.delimited);
			ffView.setColumnDelimiter(dataFileManager.getDefaultColumnDelimiter());
			ffView.setRowDelimiter(dataFileManager.getDefaultRowDelimiter());
			ffView.setNullToken(dataFileManager.getDefaultNullToken());

			File fileGen = runtimeSupport.getGeneratedSourceDirectory("file");

			File mkdirs = new FileBuilder(fileGen).subdir("synthetic-fml").mkdirs().file();

			FileUtils.forceMkdir(mkdirs);

			File file = new File(mkdirs, ffView.getId());

			FileUtils.touch(file);

			IOUtils.writeStringToFile(file, ffView.toJsonString());

			// copy both files - source and target, into a local delimited copy
			localExpected = fileRuntimeSupport.getGeneratedDataFileForCurrentTest(op.getQualifiedName() + "-source-" + ffs.getId() + "." + assertionSource.getName() + ".delimited");

			DataFile actual = dataFileManager.loadDataFile(target, ffs);
			DataFile actualView = dataFileManager.loadDataFile(localTarget, ffView);

			applicationLog.info("Caching local copy of target " + target.getName() + " to " + localTarget.getAbsolutePath());
			dataFileManager.copyDataFile(actual, actualView);

			// before processing the source, refresh the data if necessary
			if (refreshAssertionData)
			{
				// copy from the local target, since it has the schema the source will be validated against
				applicationLog.info("Refreshing assertion data " + assertionSource.getName() + " from " + localTarget.getAbsolutePath());
				artContent.referencedByCurrentTest("assertion-refresh-file");
				IOUtils.copyFiles(localTarget, assertionSource);
			}

			DataFile expected = dataFileManager.loadDataFile(assertionSource, ffView);
			DataFile expectedView = dataFileManager.loadDataFile(localExpected, ffView);

			applicationLog.info("Caching local copy of source " + assertionSource.getName() + " to " + localExpected.getAbsolutePath());
			dataFileManager.copyDataFile(expected, expectedView);

			// set the file pointers to the new files
			DataFile actualDiffSource = dataFileManager.loadDataFile(localTarget, ffView);
			DataFile expectedDiffSource = dataFileManager.loadDataFile(localExpected, ffView);

			// compare localSource to localTarget

			// we now have a delimited local file to compare the master file to (created for convenience for debugging)
			List<FileDiff> diffResults = dataFileManager.diff(expectedDiffSource, actualDiffSource);

			String failureId = FileConstants.FAIL_FILE_DIFF;

			if (operation.getFailureId() != null)
			{
				failureId = operation.getFailureId();
			}

			if (diffResults.size() != 0)
			{
				dataFileManager.report(
					diffManager,
					op,
					failureId,
					diffResults,
					fileName,
					targetFileName
				);

				throw new TestAssertionFailure("File : " + assertionSource.getAbsolutePath() + " does not match : " + target.getAbsolutePath(), failureId);
			}
		}
		catch (IOException e)
		{
			throw new TestExecutionError("Error comparing files: " + e.toString(), FileConstants.ERR_FILE_COMPARE, e);
		}

		return ClassResponder.action_code.handled;
	}

	private void simpleCheckContents(File target, ETLTestValueObject obj, AssertRequest operation, AssertRequest.AssertionMode mode, ETLTestOperation op, VariableContext context) throws TestAssertionFailure, TestExecutionError
	{
		String failureId = mode == AssertRequest.AssertionMode.NOT_EMPTY ? FileConstants.FAIL_FILE_EMPTY : FileConstants.FAIL_FILE_NOT_EMPTY;

		if (operation.getFailureId() != null)
		{
			failureId = operation.getFailureId();
		}

		if (!target.exists())
		{
			throw new TestExecutionError(target.getAbsolutePath(), FileConstants.ERR_TARGET_FILE_MISSING);
		}

		try
		{
			DataFileSchema ffs = fileRuntimeSupport.locateReferenceFileSchemaForCurrentTest(operation.getFile(), null, obj);

			applicationLog.info("Resolved data file schema: " + ffs.getId());

			context.declareAndSetStringValue(FileConstants.ASSERT_FILE_SCHEMA, ffs.getId());

			// open the file and count the lines
			DataFile df = dataFileManager.loadDataFile(target, ffs);

			DataFile.FileData data = df.getFileData();

			int rows = 0;

			try
			{
				Iterator<DataFile.FileRow> it = data.iterator();

				while (it.hasNext())
				{
					it.next();
					rows++;
				}
			}
			finally
			{
				data.dispose();
			}

			switch (mode)
			{
				case EMPTY:
					if (rows != 0)
					{
						throw new TestAssertionFailure(target.getAbsolutePath() + " contains " + rows + " rows of data", failureId);
					}
					break;
				case NOT_EMPTY:
					if (rows == 0)
					{
						throw new TestAssertionFailure(target.getAbsolutePath() + " contains no data", failureId);
					}
					break;
			}
		}
		catch(IOException exc)
		{
			throw new TestExecutionError("", FileConstants.ERR_FILE_COMPARE, exc);
		}
		catch (ResourceNotFoundException exc)
		{
			// fallback - just check the file length
			switch (mode)
			{
				case EMPTY:
					if (target.length() != 0)
					{
						throw new TestAssertionFailure(target.getAbsolutePath(), failureId);
					}
					break;
				case NOT_EMPTY:
					if (target.length() == 0)
					{
						throw new TestAssertionFailure(target.getAbsolutePath(), failureId);
					}
					break;
			}
		}
	}

	public action_code process(ETLTestMethod mt, ETLTestOperation op, ETLTestValueObject parameters, VariableContext vcontext, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		return action_code.defer;
	}
}